@@ -1,6 +1,6 @@
 // Copyright Contributors to the Open Shading Language project.
 // SPDX-License-Identifier: BSD-3-Clause
-// https://github.com/imageworks/OpenShadingLanguage
+// https://github.com/AcademySoftwareFoundation/OpenShadingLanguage
 
 // clang-format off
 
@@ -223,7 +223,7 @@ public:
 // delegating access to element storage to the DualStorage base class
 template<
          class T,           // Base data type
-         int PARTIALS=1     // Number of dimentions of partial derivs
+         int PARTIALS=1     // Number of dimensions of partial derivs
         >
 class Dual : public DualStorage<T, PARTIALS> {
     static const int elements = PARTIALS+1;   // main value + partials
