@@ -18,6 +18,7 @@
 
 #pragma once
 
+#include <cstring>
 #include <memory>
 
 #include <OSL/oslversion.h>
@@ -26,6 +27,19 @@
 /////////////////////////////////////////////////////////////////////////
 // Detect which compiler and version we're using
 
+// Notes:
+//   __GNUC__ is defined for gcc and all clang varieties
+//   __clang__ is defined for all clang varieties (generic, Apple, and Intel)
+//   __apple_build_version__ is only defined for Apple clang
+//   __INTEL_COMPILER is defined only for icc
+//   __INTEL_LLVM_COMPILER is defined only for icx
+//   _MSC_VER is defined for MSVS compiler (not gcc/clang/icc even on Windows)
+//   _WIN32 is defined on Windows regardless of compiler
+//   __CUDACC__   is defined any time we are compiling a module for Cuda
+//                (both for the host pass and the device pass). "Do this
+//                when using nvcc or clang with ptx target."
+//   __CUDA_ARCH__  is only defined when doing the device pass. "Do this only
+//                for code that will actually run on the GPU."
 
 // Define OSL_GNUC_VERSION to hold an encoded gcc version (e.g. 40802 for
 // 4.8.2), or 0 if not a GCC release. N.B.: This will be 0 for clang.
@@ -39,7 +53,7 @@
 // 30402 for clang 3.4.2), or 0 if not a generic Clang release.
 // N.B. This will be 0 for the clang Apple distributes (which has different
 // version numbers entirely).
-#if defined(__clang__) && !defined(__apple_build_version__)
+#if defined(__clang__) && !defined(__apple_build_version__) && !defined(__INTEL_LLVM_COMPILER)
 #  define OSL_CLANG_VERSION (10000*__clang_major__ + 100*__clang_minor__ + __clang_patchlevel__)
 #else
 #  define OSL_CLANG_VERSION 0
@@ -48,39 +62,59 @@
 // Define OSL_APPLE_CLANG_VERSION to hold an encoded Apple Clang version
 // (e.g. 70002 for clang 7.0.2), or 0 if not an Apple Clang release.
 #if defined(__clang__) && defined(__apple_build_version__)
+#  if defined(__INTEL_LLVM_COMPILER)
+#    error Not expected for __INTEL_LLVM_COMPILER to be defined with an __apple_build_version__
+#  endif
 #  define OSL_APPLE_CLANG_VERSION (10000*__clang_major__ + 100*__clang_minor__ + __clang_patchlevel__)
 #else
 #  define OSL_APPLE_CLANG_VERSION 0
 #endif
+// The classic Intel(r) C++ Compiler on OSX may still define __clang__
+// combine with testing OSL_INTEL_CLASSIC_COMPILER_VERSION to further differentiate if needed.
 
-// Define OSL_INTEL_COMPILER to hold an encoded Intel compiler version
-// (e.g. 1900), or 0 if not an Intel compiler.
+// Define OSL_INTEL_CLASSIC_COMPILER_VERSION to hold an encoded Intel(r) C++ Compiler version
+// (e.g. 1900), or 0 if not an Intel(r) C++ Compiler.
 #if defined(__INTEL_COMPILER)
-#  define OSL_INTEL_COMPILER __INTEL_COMPILER
+#  define OSL_INTEL_CLASSIC_COMPILER_VERSION __INTEL_COMPILER
 #else
-#  define OSL_INTEL_COMPILER 0
+#  define OSL_INTEL_CLASSIC_COMPILER_VERSION 0
 #endif
 
-// Intel's compiler on OSX may still define __clang__
-// and we have need to know when using a true clang compiler
-#if !defined(__INTEL_COMPILER) && defined(__clang__)
-    #define OSL_NON_INTEL_CLANG  __clang__
+// Define OSL_INTEL_LLVM_COMPILER_VERSION to hold an encoded Intel(r) LLVM Compiler version
+// (e.g. 20220000), or 0 if not an Intel(r) LLVM Compiler.
+// Define OSL_INTEL_CLANG_VERSION to hold the encoded Clang version the
+// Intel(r) LLVM Compiler is based on (e.g. 140000),
+// or 0 if not an Intel(r) LLVM compiler.
+#if defined(__INTEL_LLVM_COMPILER)
+#  define OSL_INTEL_LLVM_COMPILER_VERSION __INTEL_LLVM_COMPILER
+#  define OSL_INTEL_CLANG_VERSION (10000*__clang_major__ + 100*__clang_minor__ + __clang_patchlevel__)
 #else
-	#define OSL_NON_INTEL_CLANG  0
+#  define OSL_INTEL_LLVM_COMPILER_VERSION 0
+#  define OSL_INTEL_CLANG_VERSION 0
 #endif
 
+// Define OSL_ANY_CLANG to 0 or 1 to indicate if any Clang based compiler is in use
+#if defined(__clang__)
+#  define OSL_ANY_CLANG 1
+#else
+#  define OSL_ANY_CLANG 0
+#endif
+
+
 // Tests for MSVS versions, always 0 if not MSVS at all.
 #if defined(_MSC_VER)
-#  if _MSC_VER < 1900
-#    error "This version of OSL is meant to work only with Visual Studio 2015 or later"
-#  endif
+#  define OSL_MSVS_VERSION       _MSC_VER
 #  define OSL_MSVS_AT_LEAST_2013 (_MSC_VER >= 1800)
 #  define OSL_MSVS_BEFORE_2013   (_MSC_VER <  1800)
 #  define OSL_MSVS_AT_LEAST_2015 (_MSC_VER >= 1900)
 #  define OSL_MSVS_BEFORE_2015   (_MSC_VER <  1900)
 #  define OSL_MSVS_AT_LEAST_2017 (_MSC_VER >= 1910)
 #  define OSL_MSVS_BEFORE_2017   (_MSC_VER <  1910)
+#  if OSL_MSVS_BEFORE_2017
+#    error "This version of OSL is meant to work only with Visual Studio 2017 or later"
+#  endif
 #else
+#  define OSL_MSVS_VERSION       0
 #  define OSL_MSVS_AT_LEAST_2013 0
 #  define OSL_MSVS_BEFORE_2013   0
 #  define OSL_MSVS_AT_LEAST_2015 0
@@ -90,50 +124,39 @@
 #endif
 
 
-// Note about __CUDACC__ versus __CUDA_ARCH__:
-//
-// __CUDACC__     defined any time we are compiling a module for Cuda
-//                (both for the host pass and the device pass). "Do this
-//                when using nvcc or clang with ptx target."
-// __CUDA_ARCH__  is only defined when doing the device pass. "Do this only
-//                for code that will actually run on the GPU."
-
 
 /////////////////////////////////////////////////////////////////////////
 // Detect which C++ standard we're using, and handy macros.
 // See https://en.cppreference.com/w/cpp/compiler_support
 //
-// Note: oslversion.h defines OSL_BUILD_CPP11 to be 1 if OSL was built
-// using C++11. In contrast, OSL_CPLUSPLUS_VERSION defined below will be set
-// to the right number for the C++ standard being compiled RIGHT NOW. These
-// two things may be the same when compiling OSL, but they may not be the
-// same if another package is compiling against OSL and using these headers
-// (OSL may be C++11 but the client package may be newer, or vice versa --
-// use these two symbols to differentiate these cases, when important).
+// Note: oslversion.h defines OSL_BUILD_CPP to be 14, 17, etc., to reflect
+// the version that OSL itself was built with. In contrast,
+// OSL_CPLUSPLUS_VERSION defined below will be set to the right number for
+// the C++ standard being compiled RIGHT NOW by whomever is parsing these
+// header files. These two things may be the same when compiling OSL, but
+// they may not be the same if another package is compiling against OSL and
+// using these headers (e.g., OSL may be C++14 but the client package may be
+// newer, or vice versa -- use these two symbols to differentiate these
+// cases, when important).
 #if (__cplusplus >= 202001L)
 #    define OSL_CPLUSPLUS_VERSION 20
-#    define OSL_CONSTEXPR14 constexpr
 #    define OSL_CONSTEXPR17 constexpr
 #    define OSL_CONSTEXPR20 constexpr
 #elif (__cplusplus >= 201703L)
 #    define OSL_CPLUSPLUS_VERSION 17
-#    define OSL_CONSTEXPR14 constexpr
 #    define OSL_CONSTEXPR17 constexpr
 #    define OSL_CONSTEXPR20 /* not constexpr before C++20 */
 #elif (__cplusplus >= 201402L) || (defined(_MSC_VER) && _MSC_VER >= 1914)
 #    define OSL_CPLUSPLUS_VERSION 14
-#    define OSL_CONSTEXPR14 constexpr
-#    define OSL_CONSTEXPR17 /* not constexpr before C++17 */
-#    define OSL_CONSTEXPR20 /* not constexpr before C++20 */
-#elif (__cplusplus >= 201103L) || _MSC_VER >= 1900
-#    define OSL_CPLUSPLUS_VERSION 11
-#    define OSL_CONSTEXPR14 /* not constexpr before C++14 */
 #    define OSL_CONSTEXPR17 /* not constexpr before C++17 */
 #    define OSL_CONSTEXPR20 /* not constexpr before C++20 */
 #else
-#    error "This version of OSL requires C++11"
+#    error "This version of OSL requires C++14 or above"
 #endif
 
+// DEPRECATED(1.12): use C++14 constexpr
+#define OSL_CONSTEXPR14 constexpr
+
 
 
 // In C++20 (and some compilers before that), __has_cpp_attribute can
@@ -211,6 +234,11 @@
     #define OSL_CLANG_ATTRIBUTE(value)
 #endif
 
+#if OSL_GNUC_VERSION
+    #define OSL_GNUC_ATTRIBUTE(value) __attribute__((value))
+#else
+    #define OSL_GNUC_ATTRIBUTE(value)
+#endif
 
 
 #ifndef OSL_DEBUG
@@ -237,6 +265,16 @@
     #define OSL_OMP_PRAGMA(aUnQuotedPragma)
 #endif
 
+#define OSL_OMP_SIMD_LOOP(...) OSL_OMP_PRAGMA(omp simd __VA_ARGS__)
+
+#if (OSL_GCCVERSION || OSL_INTEL_CLASSIC_COMPILER_VERSION || OSL_INTEL_LLVM_COMPILER_VERSION)
+#   define OSL_OMP_COMPLEX_SIMD_LOOP(...) OSL_OMP_SIMD_LOOP(__VA_ARGS__)
+#else
+    // Ignore requests to vectorize complex/nested SIMD loops for certain
+    // compilers/versions that are not currently vectorizing complex loops.
+#   define OSL_OMP_COMPLEX_SIMD_LOOP(...)
+#endif
+
 // OSL_FORCEINLINE is a function attribute that attempts to make the
 // function always inline. On many compilers regular 'inline' is only
 // advisory. Put this attribute before the function return type, just like
@@ -446,8 +484,40 @@
 #endif
 
 
+/// OSL_ALLOCA is used to allocate smallish amount of memory on the stack,
+/// equivalent of C99 type var_name[size].
+///
+/// NOTE: in a debug build, this will assert for allocations >= 1MB, which is
+/// much too big. Hopefully this will keep us from abusing alloca and having
+/// stack overflows. The rule of thumb is that it's ok to use alloca for small
+/// things of bounded size, but not for anything that could be arbitrarily
+/// big, which should instead use heap allocation.
+#if defined(__GNUC__)
+#    define OSL_ALLOCA(type, size) (assert(size < (1<<20)), (size) != 0 ? (reinterpret_cast<type*>(__builtin_alloca((size) * sizeof(type)))) : nullptr)
+#else
+#    define OSL_ALLOCA(type, size) (assert(size < (1<<20)), (size) != 0 ? (reinterpret_cast<type*>(alloca((size) * sizeof(type)))) : nullptr)
+#endif
+
 OSL_NAMESPACE_ENTER
 
+
+/// Standards-compliant bit cast of two equally sized types. This is used
+/// equivalently to C++20 std::bit_cast, but it works prior to C++20 and
+/// it has the right decorators to work with Cuda.
+/// @version 1.13.2
+template <typename To, typename From>
+OSL_FORCEINLINE OSL_HOSTDEVICE To bitcast(const From& src) noexcept {
+    // NOTE: this is the only standards compliant way of doing this type of casting,
+    // luckily the compilers we care about know how to optimize away this idiom.
+    static_assert(sizeof(From) == sizeof(To),
+                  "bit_cast must be between objects of the same size");
+    To dst;
+    memcpy((void*)&dst, &src, sizeof(From));
+    return dst;
+}
+
+
+
 #if OSL_CPLUSPLUS_VERSION >= 20
 using std::assume_aligned;
 #else
@@ -480,4 +550,6 @@ DataT* assume_aligned(DataT* ptr)
 
 #endif
 
+static constexpr int MaxSupportedSimdLaneCount = 16;
+
 OSL_NAMESPACE_EXIT
